from unittest import TestCase
import os
import doctest

class ReadMeDocTest(TestCase):
    def test__readme_doctests(self):
        readme_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "README.rst"))
        self.assertTrue(os.path.exists(readme_path))
        result = doctest.testfile(readme_path, module_relative=False)
        self.assertEquals(result.failed, 0, "%s tests failed!" % result.failed)


from forge import ForgeTestCase
import waiting

class VirtualTimeTest(ForgeTestCase):
    def setUp(self):
        super(VirtualTimeTest, self).setUp()
        self.virtual_time = 0
        self.sleeps_performed = []
        self.predicate_satisfied = False
        self.satisfy_at_time = None
        self.satisfy_after_time = None
        self.forge.replace_with(waiting.time_module, "sleep", self._sleep)
        self.forge.replace_with(waiting.time_module, "time", self._time)
        self.forge.replace_with(waiting.deadlines.time_module, "time", self._time)
    def _sleep(self, delta):
        self.sleeps_performed.append(delta)
        self.virtual_time += delta
        self.assertGreater(1000, len(self.sleeps_performed), "Infinite loop")
    def _time(self):
        return self.virtual_time
    def predicate(self):
        if self.satisfy_at_time is not None and self.satisfy_at_time == self.virtual_time:
            self.predicate_satisfied = True
        if self.satisfy_after_time is not None and self.satisfy_after_time <= self.virtual_time:
            self.predicate_satisfied = True
        return self.predicate_satisfied

class ExponentialSleepTest(VirtualTimeTest):
    def test__exponential_sleep_start_end_multiplier(self):
        self._test__exponential_sleep([1, 2, 4, 8, 16, 32, 64, 100, 100], (1, 100, 2))
    def test__exponential_sleep_start_end_multiplier_3(self):
        self._test__exponential_sleep([1, 3, 9, 27, 81, 101, 101, 101], (1, 101, 3))
    def test__exponential_sleep_start_end_no_multiplier(self):
        self._test__exponential_sleep([1, 2, 4, 8, 16, 32, 64, 128, 201], (1, 201))
    def test__exponential_sleep_start_end_no_end_with_multiplier(self):
        self._test__exponential_sleep([1, 3, 9, 27, 81], (1, None, 3))
    def test__exponential_sleep_start_end_no_end_no_multiplier(self):
        self._test__exponential_sleep([1, 2, 4, 8, 16, 32, 64, 128, 256, 512], (1, None))
    def _test__exponential_sleep(self, sleeps, sleep_seconds_arg):
        self.satisfy_after_time = sum(sleeps)
        waiting.wait(self.predicate, sleep_seconds=sleep_seconds_arg)
        self.assertEquals(sleeps, self.sleeps_performed)

class WaitingTest(VirtualTimeTest):
    def test__waiting_does_not_expire(self):
        num_tries = 9
        sleep_seconds = 10
        self.satisfy_at_time = num_tries * sleep_seconds
        waiting.wait(self.predicate, sleep_seconds=sleep_seconds, timeout_seconds=(sleep_seconds*num_tries)+1)
    def test__waiting_expires(self):
        with self.assertRaises(waiting.TimeoutExpired):
            waiting.wait(self.predicate, sleep_seconds=3, timeout_seconds=5) # timeout_seconds intentionally not divided by sleep
        self.assertEquals(self.virtual_time, 5)
    def test__iterwait_sleep_time(self):
        sleep_seconds = 5
        num_sleeps = 7
        self.satisfy_at_time = sleep_seconds * num_sleeps
        previous_time = self.virtual_time
        for retry in waiting.iterwait(self.predicate, sleep_seconds=sleep_seconds):
            self.assertEquals(self.virtual_time, previous_time)
            previous_time = self.virtual_time + sleep_seconds
            self._sleep(1.5) # shouldn't affect the sleep
        self.assertEquals(self.virtual_time, num_sleeps * sleep_seconds)

class AggregationTest(VirtualTimeTest):
    def setUp(self):
        super(AggregationTest, self).setUp()
        self.predicates = [self.forge.create_wildcard_mock() for i in range(5)]
    def test__wait_any(self):
        for p in self.predicates:
            p().and_return(False)
        self.predicates[0]().and_return(True)
        self.forge.replay()
        waiting.wait(waiting.ANY(self.predicates))
    def test__wait_any_satisfied_on_start(self):
        for index, p in enumerate(self.predicates):
            works = (index == 2)
            p().and_return(works)
            if works:
                break
        self.forge.replay()
        waiting.wait(waiting.ANY(self.predicates))
        self.assertEquals(self.virtual_time, 0)
    def test__wait_all(self):
        for iteration in range(len(self.predicates) * 2, 0, -1):
            iteration -= 1
            for index, p in reversed(list(enumerate(self.predicates))):
                if index > iteration:
                    continue
                p().and_return(index == iteration)
        self.forge.replay()
        waiting.wait(waiting.ALL(self.predicates))
        self.assertEquals(self.virtual_time, len(self.predicates) * 2 - 1)



try:
    from flux import current_timeline as time_module
except ImportError:
    import time as time_module

class Deadline(object):
    def is_expired(self):
        raise NotImplementedError() # pragma: no cover
    def get_num_seconds_remaining(self):
        return None
class Within(Deadline):
    def __init__(self, seconds):
        super(Within, self).__init__()
        self._deadline = time_module.time() + seconds
    def is_expired(self):
        return time_module.time() >= self._deadline
    def get_num_seconds_remaining(self):
        return self._deadline - time_module.time()

class Whenever(Deadline):
    def is_expired(self):
        return False

def make_deadline(seconds):
    if seconds is None:
        return Whenever()
    return Within(seconds)



#! /usr/bin/python
import subprocess
import sys
import os

def _execute(cmd):
    if 0 != subprocess.call(cmd, shell=True):
        sys.exit(-1)

if __name__ == '__main__':
    deps = [
        "nose",
        "pyforge",
    ]
    if sys.version_info < (2, 7):
        deps.append("unittest2")

    _execute("pip install --use-mirrors {0}".format(" ".join(deps)))



from .__version__ import __version__
from contextlib import contextmanager

try:
    from flux import current_timeline as time_module
except ImportError:
    import time as time_module
from .deadlines import make_deadline as _make_deadline
from .exceptions import TimeoutExpired

def wait(*args, **kwargs):
    result = _Result()
    for x in iterwait(result=result, *args, **kwargs):
        pass
    return result.result

def iterwait(predicate, timeout_seconds=None, sleep_seconds=1, result=None, waiting_for=None):
    timeout = _make_deadline(timeout_seconds)
    if result is None:
        result = _Result()
    wait_msg = "waiting for {0}".format(waiting_for if waiting_for is not None else predicate)
    sleep_generator = _get_sleep_generator(timeout, sleep_seconds)
    while True:
        result.result = predicate()
        if result.result:
            return
        if timeout.is_expired():
            raise TimeoutExpired(timeout_seconds, waiting_for)
        with _end_sleeping(next(sleep_generator)):
            yield

@contextmanager
def _end_sleeping(total_seconds):
    deadline = _make_deadline(total_seconds)
    yield
    time_module.sleep(max(0, deadline.get_num_seconds_remaining()))

class _Result(object):
    result = None

def _get_sleep_generator(timeout, sleep_seconds):
    if type(sleep_seconds) not in (tuple, list):
        sleep_seconds = (sleep_seconds, sleep_seconds, 1)
    if len(sleep_seconds) <= 2:
        sleep_seconds = (sleep_seconds[0], sleep_seconds[1], 2)
    elif len(sleep_seconds) < 2:
        sleep_seconds = (sleep_seconds[0], None, 2)
    current_sleep, end_sleep, multiplier = sleep_seconds
    while True:
        time_remaining = timeout.get_num_seconds_remaining()
        if time_remaining is not None:
            current_sleep = min(time_remaining, current_sleep)
        current_sleep = max(0, current_sleep)
        yield current_sleep
        current_sleep *= multiplier
        if end_sleep is not None:
            current_sleep = min(end_sleep, current_sleep)

class Aggregate(object):
    def __init__(self, predicates):
        super(Aggregate, self).__init__()
        self.predicates = list(predicates)
class ANY(Aggregate):
    def __call__(self):
        return any(p() for p in self.predicates)
class ALL(Aggregate):
    def __call__(self):
        for index in range(len(self.predicates), 0, -1):
            index -= 1
            if self.predicates[index]():
                self.predicates.pop(index)
        return not self.predicates



import os
import platform
import itertools
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), "waiting", "__version__.py")) as version_file:
    exec(version_file.read())

_INSTALL_REQUIREMENTS = ["pyforge"]
if platform.python_version() < '2.7':
    _INSTALL_REQUIREMENTS.append('unittest2')

setup(name="waiting",
      classifiers = [
          "Programming Language :: Python :: 2.6",
          ],
      description="Utility for waiting for stuff to happen",
      license="BSD",
      author="Rotem Yaari",
      author_email="vmalloc@gmail.com",
      version=__version__,
      packages=find_packages(exclude=["tests"]),
      install_requires=_INSTALL_REQUIREMENTS,
      scripts=[],
      namespace_packages=[]
      )



class TimeoutExpired(Exception):
    def __init__(self, timeout_seconds, msg):
        super(TimeoutExpired, self).__init__(timeout_seconds, msg)
        self._timeout_seconds = timeout_seconds
        self._msg = msg
    def __str__(self):
        return "Timeout of {0} seconds expired waiting for {1}".format(*self.args)
    def __repr__(self):
        return "{0}: {1}".format(type(self).__name__, self)

